/* unionpgm: superimpose two pgm files, or ppm file and pgm file with specifying a coloer */
/* 2008 Takeaki Uno */

#ifndef _unionpgm_c_
#define _unionpgm_c_

#ifndef SACHICA_CELL_CHAR
  #define SACHICA_CELL_CHAR unsigned char
  #define SACHICA_CELL_END 256
#endif

#include"sachica.h"
#include"pos.c"
#include"problem.c"

#define UNIONPGM_COLOR 4
#define UNIONPGM_REVERSE_X 8
#define UNIONPGM_REVERSE_Y 16
#define UNIONPGM_1ST_COLORED 32
#define UNIONPGM_BMP 64
#define UNIONPGM_MAX_COLOR 16



// &4: output to color file
// &8: reverse for x direction
// &16: reverse for y direction
// &32: input color file for file 1

/* error messages */
void UNIONPGM_error (){
  ERROR_MES = "command explanation";
  print_err (" unionpgm: merge two pgm files\n\
unionpgm [options] input-filename1 input-filename2 output-filename\n\
-c [num]: use [num]th color for second file\n\
-C: 1st file is colored (ppm file)\n\
-r: reverse 2nd file in x direction\n\
-R: reverse 2nd file in y direction\n\
-p: make bmp file\n\
# the 1st letter of input-filename cannot be '-'.\n");

  EXIT;
}

/***************************************************************/
void UNIONPGM_read_param (SACHICA *SA, int argc, char *argv[]){
  int c=1;
  PROBLEM *PP = &SA->PP;

  if ( argc < c+4 ){ UNIONPGM_error (); return; };
  while ( argv[c][0] == '-' ){
    switch (argv[c][1]){
      case 'c': PP->problem |= UNIONPGM_COLOR;
        SA->color = atoi(argv[c+1]) % UNIONPGM_MAX_COLOR; c++;
      break; case 'r': PP->problem |= UNIONPGM_REVERSE_X;
      break; case 'R': PP->problem |= UNIONPGM_REVERSE_Y;
      break; case 'C': PP->problem |= UNIONPGM_1ST_COLORED;
      break; case 'p': PP->problem |= UNIONPGM_BMP;
      break; default: goto NEXT;
    }
    c++;
    if ( argc < c+4 ){ UNIONPGM_error (); return; }
  }

  NEXT:;
  PP->input_fname = argv[c];
  PP->position_fname = argv[c+1];
  PP->output_fname = argv[c+2];
}

/*****************************************/
void UNIONPGM_scan_pgmfile (SACHICA *SA){
  char s[10000];
  fopen2 (SA->fp1, SA->PP.input_fname, "r", EXIT);
  fopen2 (SA->fp2, SA->PP.position_fname, "r", {fclose(SA->fp1);EXIT;});

  fscanf (SA->fp1, "%s\n", s);
  fscanf (SA->fp1, "#result of comparison\n");
  fscanf (SA->fp1, "%d %d\n", &SA->xmax, &SA->ymax);
  fscanf (SA->fp1, "255\n");
  if ( strchr (s, '6') ) SA->PP.problem |= UNIONPGM_1ST_COLORED;
  
  fscanf (SA->fp2, "%s\n", s);
  fscanf (SA->fp2, "#result of comparison\n");
  fscanf (SA->fp2, "%d %d\n", &SA->xmax2, &SA->ymax2);
  fscanf (SA->fp2, "255\n");
//  if ( strchr ( s, '6' ) ) SACHICA_type2 = 1;

  SA->pxmax = MAX(SA->xmax, SA->xmax2);
  SA->pymax = MAX(SA->ymax, SA->ymax2);
}


SACHICA_CELL_CHAR SACHICA_charplus (SACHICA_CELL_CHAR c1, SACHICA_CELL_CHAR c2){
  if ( ((int)c1) + ((int)c2) > 255 ) return (255);
  else return (c1+c2);
}

/***************************************************************/
void UNIONPGM_read_pgm (SACHICA *SA, FILE *fp, int xx, int yy){
  PROBLEM *PP = &SA->PP;
  int x, y, tx, ty;
  unsigned char c1, c2=0, c3=0;
  FLOOP (y, 0, yy){  // loops for pixel
    FLOOP (x, 0, xx){
      c1 = getc (fp);
      if ( PP->problem & UNIONPGM_1ST_COLORED ){ // read color file
        c2 = getc (fp);
        c3 = getc (fp);
      } else if ( PP->problem & (UNIONPGM_COLOR + UNIONPGM_1ST_COLORED) ){  // change gray scale to color 
        c2 = ((int)c1)*SACHICA_col2[SA->color]/2;
        c3 = ((int)c1)*SACHICA_col3[SA->color]/2;
        c1 = ((int)c1)*SACHICA_col1[SA->color]/2;
      }

      tx = x; ty = y;// for reverse options
      if ( PP->problem & UNIONPGM_REVERSE_X ) tx = SA->pxmax-x-1;
      if ( PP->problem & UNIONPGM_REVERSE_Y ) ty = SA->pymax-y-1;
      
      if ( PP->problem & (UNIONPGM_COLOR + UNIONPGM_1ST_COLORED) ){ // write to color file
        SA->cell[ty][tx*3] = SACHICA_charplus (SA->cell[ty][tx*3], c1);
        SA->cell[ty][tx*3+1] = SACHICA_charplus (SA->cell[ty][tx*3+1], c2);
        SA->cell[ty][tx*3+2] = SACHICA_charplus (SA->cell[ty][tx*3+2], c3);
         // else write to gray scale file
      } else SA->cell[ty][tx] = SACHICA_charplus (SA->cell[ty][tx], c1);
    }
  }
}

/* initialization */
void UNIONPGM_init (SACHICA *SA){
  if ( SA->PP.problem & (UNIONPGM_COLOR+UNIONPGM_1ST_COLORED) ) SA->pxmax2 = SA->pxmax * 3;
  else SA->pxmax2 = SA->pxmax;

    // initialize matrix
  calloc2d (SA->cell, SA->pymax, SA->pxmax2, EXIT);
}

/* termination */
void UNIONPGM_end (SACHICA *SA){
  free2d (SA->cell);
  fclose (SA->fp1);
  fclose (SA->fp2);
  PROBLEM_end (&SA->PP);
}

/* write a pgm file */
void UNIONPGM_output (SACHICA *SA){
  int y, x;
  SACHICA_CELL_CHAR c;
  FILE2 fp;
  FILE2_open (fp, SA->PP.output_fname, "wb", EXIT);
  if ( SA->PP.problem & (UNIONPGM_COLOR+UNIONPGM_1ST_COLORED) ) FILE2_puts (&fp, "P6\n"); // color file
  else FILE2_puts (&fp, "P5\n");  // gray scale file
  FILE2_puts (&fp, "#homologs\n");
  FILE2_print_int (&fp, SA->pxmax, 0);
  FILE2_print_int (&fp, SA->pymax, ' ');
  FILE2_puts (&fp, "\n255\n");
  FLOOP (y, 0, SA->pymax){
    FLOOP (x, 0, SA->pxmax){ c=SA->cell[y][x]; FILE2_putc (&fp, (char)c); }
    FILE2_flush (&fp);
  }
  FILE2_closew (&fp);
}

/* write a bmp file */
void UNIONPGM_bmp_output (SACHICA *SA){
  int y, x, xx, flag=1;
  unsigned char *buf;
  FILE *fp;
  BMPFILEHEADER fileHeader;
  BMPINFOHEADER infoHeader;

  if ( SA->PP.problem & (UNIONPGM_COLOR+UNIONPGM_1ST_COLORED) ) flag = 3;   // color file

  fopen2 (fp, SA->PP.output_fname, "r", EXIT);
  malloc2 (buf, SA->pxmax*4, {fclose(fp);EXIT;});

  fileHeader.bfType='M'*256+'B';  // means 'BM'
  fileHeader.bfSize=sizeof(BMPFILEHEADER)+sizeof(BMPINFOHEADER)+4*SA->pxmax*SA->pymax;
  fileHeader.bfReserved1=0;
  fileHeader.bfReserved2=0;
  fileHeader.bfOffBits=sizeof(BMPFILEHEADER)+sizeof(BMPINFOHEADER);
  fwrite(&fileHeader,1,sizeof(BMPFILEHEADER),fp);
  
  infoHeader.biSize=sizeof(BMPINFOHEADER);
  infoHeader.biWidth=SA->pxmax;
  infoHeader.biHeight=SA->pymax;
  infoHeader.biPlanes=1;
  infoHeader.biBitCount=32;
  infoHeader.biCompression=0;
  infoHeader.biSizeImage=4*SA->pxmax*SA->pymax;
  infoHeader.biXPixPerMeter=3780; // 96dpi
  infoHeader.biYPixPerMeter=3780;
  infoHeader.biClrUsed=0;
  infoHeader.biClrImportant=0;

  fwrite(&infoHeader,1,sizeof(BMPINFOHEADER),fp);

  BLOOP (y, SA->pymax, 0){
    for ( x=xx=0 ; xx<SA->pxmax ; xx++,x+=flag ){
      buf[xx*4+0] = (unsigned char)SA->cell[y][x];
      buf[xx*4+1] = (unsigned char)(flag==3? SA->cell[y][x+1]: SA->cell[y][x]);
      buf[xx*4+2] = (unsigned char)(flag==3? SA->cell[y][x+2]: SA->cell[y][x]);
      buf[xx*4+3] = 0;
    }
    fwrite (buf, 1, SA->pxmax*4, fp);   // write 1 line to the file
  }
  fclose2 (fp);
}


/* main routine */
int UNIONPGM_main (int argc, char *argv[]){
  SACHICA SA;
  int problem, col2;
  
  PROBLEM_init (&SA.PP);
  UNIONPGM_read_param (&SA, argc, argv);
  UNIONPGM_scan_pgmfile (&SA);
  UNIONPGM_init (&SA);
  
  col2 = SA.color;  problem = SA.PP.problem;
  print_err ("unionpgm: x:%d, y:%d color=%d\n", SA.pxmax, SA.pymax, SA.color);

  SA.color = 0; SA.PP.problem = 0;
  UNIONPGM_read_pgm (&SA, SA.fp1, SA.xmax, SA.ymax);

  SA.color = col2; SA.PP.problem = problem;
  UNIONPGM_read_pgm (&SA, SA.fp2, SA.xmax2, SA.ymax2);

  if ( SA.PP.problem & UNIONPGM_BMP ) UNIONPGM_bmp_output (&SA);
  else UNIONPGM_output (&SA);
  UNIONPGM_end (&SA);

  return (0);
}

/*******************************************************************************/
#ifndef _NO_MAIN_
#define _NO_MAIN_
int main (int argc, char *argv[]){
  return ( UNIONPGM_main (argc, argv));
}
#endif
/*******************************************************************************/

#endif





